import os
import json

def get_project_id():
    credentials_path = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')
    if not credentials_path:
        return 'UNDEFINED'
    with open(credentials_path) as fp:
        credentials = json.load(fp)
    return credentials.get('project_id')



import logging
from google.api_core.exceptions import AlreadyExists
from google.cloud import pubsub_v1

from .tools import get_project_id


logger = logging.getLogger(__name__)


class Subscriber(object):
    def __init__(self, topic_name, name, callback, gcloud_project_id=None):
        self.subscriber_name = "{}__{}".format(topic_name, name)
        self.gcloud_project_id = gcloud_project_id or get_project_id()
        self.callback = callback
        self.subscriber = pubsub_v1.SubscriberClient()
        self.subscription_path = self.subscriber.subscription_path(  #pylint: disable=E1101
            self.gcloud_project_id, self.subscriber_name
        )
        self._ensure_subscription(topic_name)
        self.future = self.subscriber.subscribe(self.subscription_path, callback=callback)

    def _create_subscription(self, topic_name):
        topic_path = self.subscriber.topic_path(self.gcloud_project_id, topic_name)  #pylint: disable=E1101

        subscription = self.subscriber.create_subscription(  #pylint: disable=E1101
            self.subscription_path, topic_path)

        print('Subscription created: {}'.format(subscription))

    def _ensure_subscription(self, topic_name):
        try:
            self._create_subscription(topic_name)
        except AlreadyExists:
            pass



import logging

from google.auth.exceptions import DefaultCredentialsError
from gces import EventPublisher as _EventPublisher


logger = logging.getLogger(__name__)


class DisabledPublisher(object):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        pass

    def __getattr__(self, item):
        return DisabledPublisher()


class EventPublisher(object):
    def setup(self, topic_name, disabled=False):
        self._pub = _EventPublisher(topic_name) if not disabled else DisabledPublisher()


    def __getattr__(self, item):
        return getattr(self._pub, item)

class NamelessTopicError(Exception):
    """Raised when configuration topic name is None"""

event_publisher = EventPublisher()

def to_bool(value):
    if isinstance(value, bool):
        return value
    return value in ['true', 'True', 'TRUE']

def includeme(config):  # pragma: no cover
    settings = config.registry.settings
    # Prevents the creation of a topic with name 'None'
    topic_name = settings['gces.event_publisher.topic_name']
    fallback = to_bool(settings.get('gces.ext.pyramid.event_publisher.fallback_to_disabled') or False)
    if not topic_name:
        error_msg = 'Add a name to "gces.event_publisher.topic_name" in your configuration environment'
        raise NamelessTopicError(error_msg)
    try:
        event_publisher.setup(
            topic_name,
            to_bool(settings['gces.ext.pyramid.event_publisher.disabled'])
        )
    except DefaultCredentialsError as exc:
        if fallback:
            event_publisher.setup(
                topic_name,
                disabled=True
            )
            logger.error('No Google Credentials envvar Found. Using DisabledPublisher for topic "%s"', topic_name)
        else:
            raise exc



# Credits to Abhinav Upadhyay (https://github.com/abhinav-upadhyay)
import json
import datetime
from dateutil.parser import parse

class DateDateTimeDecoder(json.JSONDecoder):

    def __init__(self, *args, **kargs):
        super(DateDateTimeDecoder, self).__init__(object_hook=self.dict_to_object,
                             *args, **kargs)
    
    def dict_to_object(self, d): 
        if '__type__' not in d:
            return d

        try:
            if d['__type__'] in ['datetime', 'date']:
                dateobj = parse(d['isoformat'])
                type_ = d.pop('__type__')
                return dateobj.date() if type_ == 'date' else dateobj
            else:
                return d
        except:
            return d

class DateDateTimeEncoder(json.JSONEncoder):
    def default(self, obj): # pylint: disable=E0202
        if isinstance(obj, datetime.datetime):
            return {
                '__type__' : 'datetime',
                'isoformat' : obj.isoformat(),
            }
        elif isinstance(obj, datetime.date):
            return {
                '__type__' : 'date',
                'isoformat' : obj.isoformat(),
            }
        else:
            return json.JSONEncoder.default(self, obj)


def serialize(data):
    return json.dumps(data,cls=DateDateTimeEncoder)


def deserialize(data):
    return json.loads(data, cls=DateDateTimeDecoder)



import logging
from google.api_core.exceptions import AlreadyExists
from google.cloud import pubsub_v1

from .tools import get_project_id


logger = logging.getLogger(__name__)


class Publisher(object):
    def __init__(self, topic_name, gcloud_project_id=None):
        self.publisher = pubsub_v1.PublisherClient()
        self.gcloud_project_id = gcloud_project_id or get_project_id()
        self.topic_name = topic_name
        self.topic_path = self.publisher.topic_path(  #pylint: disable=E1101
            self.gcloud_project_id, self.topic_name)
        self._ensure_topic_exists()

    def _ensure_topic_exists(self):
        """Create a new Pub/Sub topic."""
        publisher = pubsub_v1.PublisherClient()
        try:
            topic = publisher.create_topic(self.topic_path)  #pylint: disable=E1101
        except AlreadyExists:
            topic = None

        log_msg = 'Topic "{}" Created.' if topic else 'Topic "{}" already exists.'
        log_msg = log_msg.format(self.topic_path)
        logger.info(log_msg)

    def pub(self, message):
        return self.publisher.publish(
            self.topic_path, data=message
        )



from .serializer import serialize, deserialize


class Message(object):
    def __init__(self, message):
        self.message = message
        message = deserialize(message.data)
        self.id = self.message.message_id
        self.event = message['event']
        self.data = message.get('data', {})

    def ack(self):
        self.message.ack()

class Event(object):
    def __init__(self, event, data):
        self.body = serialize(
            {
                'event': event,
                'data': data
            }
        ).encode('utf-8')



import logging
from collections import namedtuple

from . subscriber import Subscriber
from .message import Message


logger = logging.getLogger(__name__)

RegistryEntry = namedtuple('RegistryEntry', ['function', 'type', 'ack_late'])

class EventSubscriber(object):
    def __init__(self, topic_name, name, gcloud_project_id=None):
        self.gcloud_project_id = gcloud_project_id
        self.topic_name = topic_name
        self.name = name
        self._register = {}
        self._obj_desc = "{} {} ({})".format(
            self.__class__.__name__,
            name,
            topic_name
        )

        logger.info('{} Created.'.format(self._obj_desc))

    def start(self):
        self.subscriber = Subscriber(
            self.topic_name, self.name, self._callback, self.gcloud_project_id
        )
        logger.info('{} Started.'.format(self._obj_desc))


    def _callback(self, message):
        message = Message(message)
        reg = self._register.get(message.event)
        if reg:
            if not reg.ack_late:
                message.ack()

            if reg.type == 'function':
                reg.function(message.data)
            elif reg.type == 'task':
                reg.function.delay(message.data)
            logger.info('Event {} processed with "{}" callback.'.format(message.event, reg.function.__name__))
        else:
            logger.info('Event {} Ignored.'.format(message.event))
        message.ack()

    def register_fsub(self, event_name, function, ack_late=True):
        self._register[event_name] = RegistryEntry(
            function=function,
            ack_late=ack_late,
            type='function'
        )
        logger.info(
            '{}: Event listener function "{}" registered for {}.'.format(
                self._obj_desc,
                function.__name__,
                event_name
            )
        )

    def register_tsub(self, event_name, function, ack_late=True):
        self._register[event_name] = RegistryEntry(
            function=function,
            ack_late=ack_late,
            type='task'
        )
        logger.info(
            '{}: Event listener task "{}" registered for {}.'.format(
                self._obj_desc,
                function.__name__,
                event_name
            )
        )



import logging

from .publisher import Publisher
from .message import Event


logger = logging.getLogger(__name__)


class EventPublisher(object):
    def __init__(self, topic_name, gcloud_project_id=None):
        self.publisher = Publisher(topic_name, gcloud_project_id)
        self._obj_desc = '{} "{}"'.format(
            self.__class__.__name__,
            topic_name
        )
        logger.info('{} Created'.format(self._obj_desc))

    def notify(self, event_name, data):
        event = Event(event_name, data)
        pub = self.publisher.pub(event.body)
        logger.info(
            '{}: Event {} notification published.'.format(
                self._obj_desc, event_name
            )
        )
        return pub



# -*- encoding: utf-8 -*-
import celery
from celery.utils.log import get_task_logger
from kombu import Exchange, Queue

from minicache import cache

from gces import EventPublisher


logger = get_task_logger(__name__)


GCES_PUBLISHER_QUEUE = '__gces__queue__.__dispatcher__'
GCES_PUBLISHER_NAME = '__gces_dispatch__'

class EventPublisherTask(object):
    def __init__(self, topic_name):
        self.topic_name = topic_name

    def notify(self, event_name, data):
        _event_publisher_task.delay(self.topic_name, event_name, data)



def _send_notification(topic_name, event_name, data):
    epub = cache.get(topic_name) or EventPublisher(topic_name)
    cache.update(topic_name, epub)
    return epub.notify(event_name, data)



@celery.shared_task(bind=True, name=GCES_PUBLISHER_NAME, queue=GCES_PUBLISHER_QUEUE, acks_late=True, ignore_result=True)
def _event_publisher_task(self, topic_name, event_name, data):
    try:
        return _send_notification(topic_name, event_name, data)
    except Exception as exc:
        cache.clear(topic_name)
        try:
            raise self.retry(exc)
        except celery.exceptions.MaxRetriesExceededError:
            # Send event to a dead letter queue.
            _event_publisher_task.apply_async(
                (topic_name, event_name, data),
                queue='{}.__deadletter__'.format(GCES_PUBLISHER_QUEUE)
            )


def _register_celery_3(celery_app, queue):
    if celery_app.conf.CELERY_QUEUES:
        celery_app.conf.CELERY_QUEUES.append(queue)
    else:
        celery_app.conf.CELERY_QUEUES = [queue]


def _register_celery_4(celery_app, queue):
    if celery_app.conf.task_queues:
        celery_app.conf.task_queues.append(queue)
    else:
        celery_app.conf.task_queues = [queue]


def register_publisher(celery_app):
    queue = Queue(
        GCES_PUBLISHER_QUEUE,
        Exchange(GCES_PUBLISHER_QUEUE),
        routing_key=GCES_PUBLISHER_QUEUE
    )
    if not hasattr(celery_app.conf, 'task_queues'):  # Celery 3
        _register_celery_3(celery_app=celery_app, queue=queue)
    else:                                            # Celery 4
        _register_celery_4(celery_app=celery_app, queue=queue)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from setuptools import setup


def long_desc_img_replacer(long_desc):
    replacements = (
        (
            '(docs/overview.svg)',
            '(https://raw.githubusercontent.com/debonzi/gc-event-system/master/docs/overview.svg?sanitize=true)'
        ),
    )
    for f, t in replacements:
        long_desc = long_desc.replace(f, t)
    return long_desc


with open('README.md') as f:
    long_description = long_desc_img_replacer(f.read())

requires = [
    'google-cloud-pubsub',
    'python-dateutil'
]

extras_require = {
    'celery': [
        'celery[redis]>=3.1.20',
        'minicache==0.0.1',
    ],
    'test': [
        'coverage==4.5.1',
        'pytest==3.8.1',
        'pytest-cov==2.6.0',
        'mock==2.0.0',
    ],
    'ci': [
        'python-coveralls==2.9.1',
    ]
}


setup(name='gces',
      version='0.0.12-alpha',
      description='Google Cloud Event System.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Daniel Debonzi',
      author_email='debonzi@gmail.com',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6'
      ],
      install_requires=requires,
      extras_require=extras_require,
      url='https://github.com/debonzi/gc-event-system',
      packages=[
          'gces',
          'gces.ext',
          'gces.ext.celery',
          'gces.ext.pyramid'
      ],
      )
